# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np

from jpu.registry import UnitRegistry


def test_tree_flatten():
    u = UnitRegistry()
    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.m

    val, _ = jax.tree_flatten(q)
    assert len(val) == 1
    np.testing.assert_allclose(val[0], x)


def test_jittable():
    u = UnitRegistry()
    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.m

    @jax.jit
    def func(q):
        assert q.u == u.m
        return q + 4.5 * u.km

    res = func(q)
    assert res.units == u.m
    np.testing.assert_allclose(res.magnitude, x + 4500.0)


def test_ducktype():
    u = UnitRegistry()
    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.m

    res = q.sum()
    assert res.units == u.m
    np.testing.assert_allclose(res.magnitude, x.sum())

    @jax.jit
    def func(q):
        return q.sum()

    res = func(q)
    assert res.units == u.m
    np.testing.assert_allclose(res.magnitude, x.sum())


def test_unary_ops():
    u = UnitRegistry()

    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.m

    for func in [
        lambda q: q**2,
        lambda q: q.sum(),
        lambda q: 2 * q,
    ]:
        res = func(q)
        np.testing.assert_allclose(res.magnitude, func(x))
        res = jax.jit(func)(q)
        np.testing.assert_allclose(res.magnitude, func(x))



# mypy: ignore-errors


def test_readme():
    import jax
    import numpy as np

    from jpu import UnitRegistry
    from jpu import numpy as jnpu

    u = UnitRegistry()

    @jax.jit
    def projectile_motion(v_init, theta, time, g=u.standard_gravity):
        """Compute the motion of a projectile with support for units"""
        x = v_init * time * jnpu.cos(theta)
        y = v_init * time * jnpu.sin(theta) - 0.5 * g * jnpu.square(time)
        return x.to(u.m), y.to(u.m)

    x, y = projectile_motion(
        5.0 * u.km / u.h, 60 * u.deg, np.linspace(0, 1, 50) * u.s
    )
    assert x.units == u.m
    assert y.units == u.m



__all__ = ["UnitRegistry"]

from typing import Any, Tuple

import jax
from jax.tree_util import register_pytree_node
from pint import UnitRegistry as PintUnitRegistry
from pint import compat


class UnitRegistry(PintUnitRegistry):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        # Register the Quantity produced by this registry with JAX
        def flatten_quantity(q: Any) -> Tuple[Tuple[Any], Any]:
            return (q.magnitude,), q.u

        def unflatten_quantity(aux_data: Any, children: Tuple[Any]) -> Any:
            return self.Quantity(children[0], aux_data)

        register_pytree_node(
            self.Quantity, flatten_quantity, unflatten_quantity
        )


def is_duck_array_type(cls: Any) -> bool:
    return issubclass(cls, jax.core.Tracer) or _is_duck_array_type(cls)


_is_duck_array_type = compat.is_duck_array_type
compat.is_duck_array_type = is_duck_array_type



# mypy: ignore-errors

import sys
from functools import wraps
from inspect import signature
from itertools import chain

import jax.numpy as jnp
from pint import numpy_func


# Based closely on the pint implementation:
# https://github.com/hgrecco/pint/blob/37a61ede6fbd628c7dc160eb36278cf41c96484c/pint/facets/numpy/numpy_func.py#L244
def _implement_func(func_name, input_units=None, output_unit=None):
    jax_func = getattr(jnp, func_name)

    @wraps(jax_func)
    def wrapped(*args, **kwargs):
        try:
            first_input_units = numpy_func._get_first_input_units(args, kwargs)
        except TypeError:
            # None of the inputs are quantities
            return jax_func(*args, **kwargs)
        if input_units == "all_consistent":
            # Match all input args/kwargs to same units
            (
                stripped_args,
                stripped_kwargs,
            ) = numpy_func.convert_to_consistent_units(
                *args, pre_calc_units=first_input_units, **kwargs
            )
        else:
            if isinstance(input_units, str):
                # Conversion requires Unit, not str
                pre_calc_units = first_input_units._REGISTRY.parse_units(
                    input_units
                )
            else:
                pre_calc_units = input_units

            # Match all input args/kwargs to input_units, or if input_units is None,
            # simply strip units
            (
                stripped_args,
                stripped_kwargs,
            ) = numpy_func.convert_to_consistent_units(
                *args, pre_calc_units=pre_calc_units, **kwargs
            )

        # Determine result through plain numpy function on stripped arguments
        result_magnitude = jax_func(*stripped_args, **stripped_kwargs)

        if output_unit is None:
            # Short circuit and return magnitude alone
            return result_magnitude
        elif output_unit == "match_input":
            result_unit = first_input_units
        elif output_unit in [
            "sum",
            "mul",
            "delta",
            "delta,div",
            "div",
            "invdiv",
            "variance",
            "square",
            "sqrt",
            "cbrt",
            "reciprocal",
            "size",
        ]:
            result_unit = numpy_func.get_op_output_unit(
                output_unit,
                first_input_units,
                tuple(chain(args, kwargs.values())),
            )
        else:
            result_unit = output_unit

        return first_input_units._REGISTRY.Quantity(
            result_magnitude, result_unit
        )

    module = sys.modules["jpu.numpy"]
    module.__all__.append(func_name)
    setattr(module, func_name, wrapped)


def _implement_by_argname(func_name, unit_arguments, wrap_output=True):
    jax_func = getattr(jnp, func_name)

    @wraps(jax_func)
    def wrapped(*args, **kwargs):
        # Bind given arguments to the NumPy function signature
        bound_args = signature(jax_func).bind(*args, **kwargs)

        # Skip unit arguments that are supplied as None
        valid_unit_arguments = [
            label
            for label in unit_arguments
            if label in bound_args.arguments
            and bound_args.arguments[label] is not None
        ]

        # Unwrap valid unit arguments, ensure consistency, and obtain output wrapper
        (
            unwrapped_unit_args,
            output_wrap,
        ) = numpy_func.unwrap_and_wrap_consistent_units(
            *(bound_args.arguments[label] for label in valid_unit_arguments)
        )

        # Call NumPy function with updated arguments
        for i, unwrapped_unit_arg in enumerate(unwrapped_unit_args):
            bound_args.arguments[valid_unit_arguments[i]] = unwrapped_unit_arg
        ret = jax_func(*bound_args.args, **bound_args.kwargs)

        # Conditionally wrap output
        if wrap_output:
            return output_wrap(ret)
        else:
            return ret

    module = sys.modules["jpu.numpy"]
    module.__all__.append(func_name)
    setattr(module, func_name, wrapped)


def implement_numpy_functions():
    for func_name in numpy_func.strip_unit_input_output_ufuncs:
        _implement_func(func_name)

    for func_name in numpy_func.matching_input_bare_output_ufuncs:
        _implement_func(func_name, input_units="all_consistent")

    for (
        func_name,
        out_unit,
    ) in numpy_func.matching_input_set_units_output_ufuncs.items():
        _implement_func(
            func_name, input_units="all_consistent", output_unit=out_unit
        )

    for func_name, (in_unit, out_unit) in numpy_func.set_units_ufuncs.items():
        _implement_func(func_name, input_units=in_unit, output_unit=out_unit)

    for func_name in numpy_func.matching_input_copy_units_output_ufuncs:
        _implement_func(
            func_name, input_units="all_consistent", output_unit="match_input"
        )

    for func_name in numpy_func.copy_units_output_ufuncs:
        _implement_func(func_name, output_unit="match_input")

    for func_name, unit_op in numpy_func.op_units_output_ufuncs.items():
        _implement_func(func_name, output_unit=unit_op)

    for func_name in ["cumprod", "cumproduct", "nancumprod"]:
        _implement_func(func_name, input_units="", output_unit="")

    for func_name in ["block", "hstack", "vstack", "dstack", "column_stack"]:
        _implement_func(
            func_name, input_units="all_consistent", output_unit="match_input"
        )

    for func_name in [
        "size",
        "isreal",
        "iscomplex",
        "shape",
        "ones_like",
        "zeros_like",
        "empty_like",
        "argsort",
        "argmin",
        "argmax",
        "alen",
        "ndim",
        "nanargmax",
        "nanargmin",
        "count_nonzero",
        "nonzero",
        "result_type",
    ]:
        _implement_func(func_name)

    for unit_type, funcs in {
        "sum": ["std", "nanstd", "sum", "nansum", "cumsum", "nancumsum"],
        "mul": ["cross", "trapz", "dot"],
        "delta": ["diff", "ediff1d"],
        "delta,div": ["gradient"],
        "variance": ["var", "nanvar"],
    }.items():
        for func_name in funcs:
            _implement_func(func_name, output_unit=unit_type)

    for func_name, unit_arguments, wrap_output in [
        ("expand_dims", "a", True),
        ("squeeze", "a", True),
        ("rollaxis", "a", True),
        ("moveaxis", "a", True),
        ("around", "a", True),
        ("diagonal", "a", True),
        ("mean", "a", True),
        ("ptp", "a", True),
        ("ravel", "a", True),
        ("round_", "a", True),
        ("sort", "a", True),
        ("median", "a", True),
        ("nanmedian", "a", True),
        ("transpose", "a", True),
        ("copy", "a", True),
        ("average", "a", True),
        ("nanmean", "a", True),
        ("swapaxes", "a", True),
        ("nanmin", "a", True),
        ("nanmax", "a", True),
        ("percentile", "a", True),
        ("nanpercentile", "a", True),
        ("quantile", "a", True),
        ("nanquantile", "a", True),
        ("flip", "m", True),
        ("fix", "x", True),
        ("trim_zeros", ["filt"], True),
        ("broadcast_to", ["array"], True),
        ("amax", ["a", "initial"], True),
        ("amin", ["a", "initial"], True),
        ("searchsorted", ["a", "v"], False),
        ("isclose", ["a", "b"], False),
        ("nan_to_num", ["x", "nan", "posinf", "neginf"], True),
        ("clip", ["a", "a_min", "a_max"], True),
        ("append", ["arr", "values"], True),
        ("compress", "a", True),
        ("linspace", ["start", "stop"], True),
        ("tile", "A", True),
        ("rot90", "m", True),
        ("insert", ["arr", "values"], True),
        ("resize", "a", True),
        ("reshape", "a", True),
        ("allclose", ["a", "b"], False),
        ("intersect1d", ["ar1", "ar2"], True),
    ]:
        _implement_by_argname(func_name, unit_arguments, wrap_output)



# -*- coding: utf-8 -*-
"""
JAX + Units
===========
"""

__all__ = ["numpy", "UnitRegistry"]

from . import numpy
from .numpy_helper import implement_numpy_functions
from .registry import UnitRegistry

implement_numpy_functions()


__author__ = "Dan Foreman-Mackey"
__email__ = "foreman.mackey@gmail.com"
__uri__ = "https://github.com/dfm/jax_astropy_units"
__license__ = "MIT"
__description__ = "A JAX + AstroPy units mashup."
__copyright__ = "2022 Simons Foundation, Inc"
__contributors__ = (
    "https://github.com/dfm/jax_astropy_units/graphs/contributors"
)
__bibtex__ = __citation__ = """TBD"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
import re

from setuptools import find_packages, setup

# PROJECT SPECIFIC

NAME = "jpu"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", "jpu", "__init__.py")
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
]
INSTALL_REQUIRES = ["jax", "jaxlib", "pint"]
EXTRA_REQUIRE = {"test": ["pytest>=3.6"]}

# END PROJECT SPECIFIC


HERE = os.path.dirname(os.path.realpath(__file__))


def read(*parts: str) -> str:
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta: str, meta_file: str = read(META_PATH)) -> str:
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), meta_file, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


if __name__ == "__main__":
    setup(
        name=NAME,
        use_scm_version={
            "write_to": os.path.join(
                "src", NAME, "{0}_version.py".format(NAME)
            ),
            "write_to_template": '__version__ = "{version}"\n',
        },
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        url=find_meta("uri"),
        license=find_meta("license"),
        description=find_meta("description"),
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        packages=PACKAGES,
        package_dir={"": "src"},
        package_data={"jpu": ["py.typed"]},
        include_package_data=True,
        python_requires=">=3.6",
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRA_REQUIRE,
        classifiers=CLASSIFIERS,
        zip_safe=True,
    )



# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np
import pint
import pytest

import jpu.numpy as jun
from jpu import UnitRegistry


def is_quantity(q):
    return hasattr(q, "magnitude") and hasattr(q, "units")


def assert_quantity_allclose(a, b):
    if is_quantity(a):
        assert is_quantity(b)
        assert str(a.units) == str(b.units)
        np.testing.assert_allclose(a.magnitude, b.magnitude)
    else:
        assert not is_quantity(b)
        np.testing.assert_allclose(a, b)


def test_type_wrapping():
    u = UnitRegistry()
    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.kpc
    assert q.units == u.kpc
    np.testing.assert_allclose(q.magnitude, x)
    assert type(q.magnitude) == type(x)


def test_array_ops():
    u = UnitRegistry()
    x = jnp.array([1.4, 2.0, -5.9])
    q = x * u.kpc

    # Addition
    res = q + np.array(0.01) * u.Mpc
    assert res.units == u.kpc
    np.testing.assert_allclose(res.magnitude, x + 10)
    assert type(res.magnitude) == type(x)

    # Different order
    res = np.array(0.01) * u.Mpc + q
    assert res.units == u.Mpc
    np.testing.assert_allclose(res.magnitude, 1e-3 * (x + 10))
    assert type(res.magnitude) == type(x)

    # Subtraction
    res = q - np.array(0.01) * u.Mpc
    assert res.units == u.kpc
    np.testing.assert_allclose(res.magnitude, x - 10)
    assert type(res.magnitude) == type(x)

    # Multiplication
    res = 2 * q
    assert res.units == u.kpc
    np.testing.assert_allclose(res.magnitude, 2 * x)
    assert type(res.magnitude) == type(x)

    # Division
    res = q / (2 * u.kpc)
    assert res.units == u.dimensionless
    np.testing.assert_allclose(res.magnitude, 0.5 * x)
    assert type(res.magnitude) == type(x)


@pytest.mark.parametrize(
    "func,in_unit",
    [
        ("exp", [""]),
        ("log", [""]),
        ("sin", ["degree"]),
        ("sin", ["radian"]),
        ("arctan2", ["m", "m"]),
        ("arctan2", ["m", "foot"]),
        ("argsort", ["day"]),
        ("std", ["day"]),
        ("var", ["m"]),
        ("dot", ["m", "s"]),
        ("median", ["m"]),
        ("cumprod", [""]),
        ("any", ["kpc"]),
    ],
)
def test_unary(func, in_unit):
    f = (lambda x: x**2) if func == "log" else (lambda x: x)

    pu = pint.UnitRegistry()
    np_args = []
    for n, iu in enumerate(in_unit):
        x = f(np.array([1.4, 2.0, -5.9]) - n) * pu(iu)
        np_args.append(x)

    u = UnitRegistry()
    jun_args = []
    for n, iu in enumerate(in_unit):
        x = f(jnp.array([1.4, 2.0, -5.9]) - n) * u(iu)
        jun_args.append(x)

    np_func = getattr(np, func)
    np_res = np_func(*np_args)

    jun_func = getattr(jun, func)
    jun_res = jun_func(*jun_args)
    assert_quantity_allclose(jun_res, np_res)
    jun_res = jax.jit(jun_func)(*jun_args)
    assert_quantity_allclose(jun_res, np_res)

    np_res_no_units = np_func(*(x.magnitude for x in np_args))
    jun_res_no_units = jun_func(*(x.magnitude for x in jun_args))
    np.testing.assert_allclose(jun_res_no_units, np_res_no_units)



# mypy: ignore-errors

__all__ = [
    "modf",
    "frexp",
    "power",
    "add",
    "subtract",
    "meshgrid",
    "full_like",
    "interp",
    "where",
    "concatenate",
    "stack",
    "einsum",
    "any",
    "all",
]

import jax.numpy as jnp
from pint.numpy_func import (
    _get_first_input_units,
    _is_quantity,
    convert_to_consistent_units,
    get_op_output_unit,
    unwrap_and_wrap_consistent_units,
)


def modf(x, *args, **kwargs):
    (x,), output_wrap = unwrap_and_wrap_consistent_units(x)
    return tuple(output_wrap(y) for y in jnp.modf(x, *args, **kwargs))


def frexp(x, *args, **kwargs):
    (x,), output_wrap = unwrap_and_wrap_consistent_units(x)
    mantissa, exponent = jnp.frexp(x, *args, **kwargs)
    return output_wrap(mantissa), exponent


def power(x1, x2):
    if _is_quantity(x1):
        return x1**x2
    else:
        return x2.__rpow__(x1)


def add(x1, x2, *args, **kwargs):
    (x1, x2), output_wrap = unwrap_and_wrap_consistent_units(x1, x2)
    return output_wrap(jnp.add(x1, x2, *args, **kwargs))


def subtract(x1, x2, *args, **kwargs):
    (x1, x2), output_wrap = unwrap_and_wrap_consistent_units(x1, x2)
    return output_wrap(jnp.subtract(x1, x2, *args, **kwargs))


def meshgrid(*xi, **kwargs):
    # Simply need to map input units to onto list of outputs
    input_units = (x.units for x in xi)
    res = jnp.meshgrid(*(x.m for x in xi), **kwargs)
    return [out * unit for out, unit in zip(res, input_units)]


def full_like(a, fill_value, **kwargs):
    # Make full_like by multiplying with array from ones_like in a
    # non-multiplicative-unit-safe way
    if hasattr(fill_value, "_REGISTRY"):
        return fill_value._REGISTRY.Quantity(
            (jnp.ones_like(a, **kwargs) * fill_value.m),
            fill_value.units,
        )
    else:
        return jnp.ones_like(a, **kwargs) * fill_value


def interp(x, xp, fp, left=None, right=None, period=None):
    # Need to handle x and y units separately
    (x, xp, period), _ = unwrap_and_wrap_consistent_units(x, xp, period)
    (fp, right, left), output_wrap = unwrap_and_wrap_consistent_units(
        fp, left, right
    )
    return output_wrap(
        jnp.interp(x, xp, fp, left=left, right=right, period=period)
    )


def where(condition, *args):
    args, output_wrap = unwrap_and_wrap_consistent_units(*args)
    return output_wrap(jnp.where(condition, *args))


def concatenate(sequence, *args, **kwargs):
    sequence, output_wrap = unwrap_and_wrap_consistent_units(*sequence)
    return output_wrap(jnp.concatenate(sequence, *args, **kwargs))


def stack(arrays, *args, **kwargs):
    arrays, output_wrap = unwrap_and_wrap_consistent_units(*arrays)
    return output_wrap(jnp.stack(arrays, *args, **kwargs))


def einsum(subscripts, *operands, **kwargs):
    if not any(_is_quantity(x) for x in operands):
        return jnp.einsum(subscripts, *operands, **kwargs)
    operand_magnitudes, _ = convert_to_consistent_units(
        *operands, pre_calc_units=None
    )
    output_unit = get_op_output_unit(
        "mul", _get_first_input_units(operands), operands
    )
    return jnp.einsum(subscripts, *operand_magnitudes, **kwargs) * output_unit


def any(a, *args, **kwargs):
    if not _is_quantity(a):
        return jnp.any(a, *args, **kwargs)
    if a._is_multiplicative:
        return jnp.any(a._magnitude, *args, **kwargs)
    else:
        raise ValueError(
            "Boolean value of Quantity with offset unit is ambiguous."
        )


def all(a, *args, **kwargs):
    if not _is_quantity(a):
        return jnp.all(a, *args, **kwargs)
    if a._is_multiplicative:
        return jnp.all(a._magnitude, *args, **kwargs)
    else:
        raise ValueError(
            "Boolean value of Quantity with offset unit is ambiguous."
        )
